/*
 * Copyright 2014-2024  the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package dev.jeka.core.tool.builtins.tooling.ide;

import dev.jeka.core.api.depmanagement.JkDependencySet;
import dev.jeka.core.api.file.JkPathSequence;
import dev.jeka.core.api.java.JkClassLoader;
import dev.jeka.core.api.java.JkUrlClassLoader;
import dev.jeka.core.api.project.JkIdeSupport;
import dev.jeka.core.api.project.JkIdeSupportSupplier;
import dev.jeka.core.api.system.JkLog;
import dev.jeka.core.tool.JkRunbase;

import java.net.URLClassLoader;
import java.nio.file.Path;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

final class IdeSupport {

    private IdeSupport() {
    }

    static JkIdeSupport getProjectIde(JkRunbase runbase) {
        List<JkIdeSupportSupplier> suppliers = runbase.getBeans().stream()
                .filter(JkIdeSupportSupplier.class::isInstance)
                .map(JkIdeSupportSupplier.class::cast)
                .collect(Collectors.toList());
        return suppliers.stream()
                .map(supplier -> {
                    JkIdeSupport ideSupport = supplier.getJavaIdeSupport();
                    if (ideSupport != null) {
                        JkLog.verbose("Use %s class as IDE support supplier.", supplier.getClass().getName());
                    }
                    return ideSupport;
                })
                .filter(Objects::nonNull)
                .findFirst().orElse(null);
    }

    static JkDependencySet classpathAsDependencySet() {
        JkClassLoader classLoader = JkClassLoader.ofCurrent();
        final JkPathSequence classpath;
        if (classLoader.get() instanceof URLClassLoader) {
            classpath = JkUrlClassLoader.ofCurrent().getDirectClasspath();
        } else {
            classpath = JkPathSequence.ofSysPropClassPath();
        }
        JkDependencySet dependencySet = JkDependencySet.of();
        for (Path entry : classpath) {
            dependencySet = dependencySet.andFiles(entry);
        }
        return dependencySet;
    }
}
